################################################################################
#
#  open class REdefinitions for built-in ruby types to add quantifiers
#
################################################################################

class Enumerable::Enumerator
  def in_parallel(&block)
    backup = $update_store

    stuff = []

    self.each do |item|
      $update_store = []
      block.call(item)
      stuff << Exp_Parallel.new($update_store)
    end

    $update_store = backup + stuff
  end
end

class Integer
  def self.exists(&block)
    quantify(:exists, SymbInteger, block)
  end

  def self.forall(&block)
    quantify(:forall, SymbInteger, block)
  end
end

class String
  def self.exists(&block)
    quantify(:exists, SymbString, block)
  end

  def self.forall(&block)
    quantify(:forall, SymbString, block)
  end
end

#todo: implement
class Sequence
  def self.exists(&block)
    quantify(:exists, SymbSeq, block)
  end

  def self.forall(&block)
    quantify(:forall, SymbSeq, block)
  end
end

class Array
  alias :old_each :each
  alias :old_include :include?

  def forall(&block)
    if $symbolic_execution then
      quantify(:forall, SymbObject, block)
    else
      self.each do |e|
        if !block.call(e) then
          return false
        end
      end

      return true
    end
  end

  def exists(&block)
    if $symbolic_execution then
      quantify(:exists, SymbObject, block)
    else
      self.each do |e|
        puts e
        if block.call(e) then
          return e
        end
      end

      return false
    end
  end

  def other_each(&block)
    if $symbolic_execution then 
      Exp_Call.new(:each, [self, block.call(SymbObject.new)])
    else
      old_each &block
    end
  end
end


# create "instruction store" globally, ALL side effects go in there!!!
# class Hash
#   alias :old_init :initialize
#   alias :old_store :store
#   alias :old_delete :delete
#   alias :old_equals :==

#   @@gensym = 1

#   def initialize
#     @name = "hash_var_" + @@gensym.to_s
#     @@gensym = @@gensym + 1
#     old_init
#   end

#   def []=(k, v)
#     if $symbolic_execution then
#       $update_store << Exp_Call.new(:update_hash, [self, k, v])
#     else
#       old_store(k, v)
#     end
#   end

#   def delete(k)
#     if $symbolic_execution then
#       $update_store << Exp_Call.new(:delete_hash, [self, k])
#     else
#       old_delete(k)
#     end
#   end

#   def ==(arg)
#     if $symbolic_execution then
#       Exp_Call.new(:equals, [self, arg])
#     else
#       old_equals(arg)
#     end
#   end


#   def to_s
#     @name
#   end

# end



# ********************************************************************************
# randint -- this is clever
# ********************************************************************************

def randInt(a, b)
  RandIntObj.new(a, b)
end

class RandIntObj
  @@num = 0

  def initialize(a,b)
    @@num = @@num + 1
    @mynum = @@num
    @a = a
    @b = b
  end

  def get_a
    @a
  end

  def get_b
    @b
  end

  def to_s
    if !($objects_used.include? self) then
      $objects_used << self
    end

    "RandInt" + @mynum.to_s
  end

  def times &block
    old_update_store = $update_store
    $update_store = []
    block.call # for the side effect
    body = $update_store
    $update_store = old_update_store

    $update_store << Exp_Var.new(:loop_index, 0)

    $update_store << Exp_Loop.new(Exp_Call.new(:not_equals, [:loop_index, self]),
                                  [Exp_Update.new(:loop_index, Exp_Call.new(:+, [:loop_index, 1]))] + body)
  end
end
