################################################################################
#
#  class definitions for the expression types of rubicon's target language
#
################################################################################

class Exp_Call
  alias :old_equals :==

  def initialize(rator, rands)
    @rator = rator
    @rands = rands

    rands.each do |r|
      if r.class == String and !$strings.include? r then
        $strings << r
      end
    end

    # rands.each do |r|
    #   symbolic r
    # end
  end

  def blank=(arg)
    # this is pretty nasty!
    if @rands[0].instance_variable_get(:@is_blank) == nil then
      @rands[0].instance_variable_set(:@is_blank, Hash.new)
    end

    @rands[0].instance_variable_get(:@is_blank)[@rands[1]] = arg
  end

  def blank?
    if @rands[0].instance_variable_get(:@is_blank) == nil then
      nil
    else
      @rands[0].instance_variable_get(:@is_blank)[@rands[1]]
    end
  end

  def ==(arg)
    Exp_Call.new(:equals, [self, arg])
  end

  def &(arg)
    Exp_And.new(self, arg)
  end

  def |(arg)
    Exp_Or.new(self, arg)
  end

  def implies(&block)
    Exp_Implies.new(self, block.call)
  end

  def method_missing(meth, *args, &block)
    #args
    # puts "method missing " + meth.to_s
    # puts "self is " + self.to_s

    if meth.old_equals(:+) then
      Exp_Call.new(meth, [self] + args)
    elsif meth.old_equals(:map) then
      block.call(self)
    else
      Exp_Call.new(:field_get, [self, meth])
    end
  end

  def to_s
    def make_field_reference
      # if $defining_objects then
      #   $fields[@rands[1].to_s][$objects_map[@rands[0].to_s]]
      # else
      #   @rands[1].to_s + "[" + @rands[0].to_s + "]"
      # end

      @rands[0].to_s + "." + @rands[1].to_s
    end

    if @rator.old_equals(:field_get) then
      make_field_reference
    elsif @rator.old_equals(:field_set) then
      make_field_reference + " = " + @rands[2].to_s
    elsif @rator.old_equals(:equals) then
      "(" + @rands[0].to_s + " == " + @rands[1].to_s + ")"
    elsif @rator.old_equals(:not_equals) then
      "(" + @rands[0].to_s + " != " + @rands[1].to_s + ")"
    elsif @rator.old_equals(:include) then
      @rands[1].to_s + " in " + @rands[0].to_s
    elsif [:+, :-, :==].include? @rator then
      "(" + @rands[0].to_s + " " + @rator.to_s + " " + @rands[1].to_s + ")"
    elsif @rator.old_equals(:skip) then
      "Skip"
    elsif @rator.old_equals(:should_not) then
      "!(" + Exp_Call.new(:should, @rands).to_s + ")"
    elsif @rator.old_equals(:should) then
      if @rands[1].old_equals(:include) then
        @rands[2].to_s + " in " + @rands[0].to_s
      elsif @rands[1].old_equals(:==) then
        @rands[2].to_s + " = " + @rands[0].to_s
      else
        "call(" + @rator.to_s + ", " + @rands.join(", ") + ")"        
      end
    else
      "call(" + @rator.to_s + ", " + @rands.join(", ") + ")"
    end
  end

end

class Exp_Var
  def initialize(var, val)
    @var = var
    @val = val
  end

  def to_s
    "var " + @var.to_s + " = " + @val.to_s
  end
end

class Exp_Update
  def initialize(var, val)
    @var = var
    @val = val
  end

  def to_s
    @var.to_s + " = " + @val.to_s
  end
end

class Exp_Not
  def initialize(e)
    @e = e
  end

  def &(arg)
    Exp_And.new(self, arg)
  end

  def |(arg)
    Exp_Or.new(self, arg)
  end

  def to_s
    "!" + @e.to_s
  end
end


class Exp_And
  def initialize(a,b)
    @a = a
    @b = b
  end

  def &(arg)
    Exp_And.new(self, arg)
  end

  def |(arg)
    Exp_Or.new(self, arg)
  end

  def implies(&block)
    Exp_Implies.new(self, block.call)
  end

  def to_s
    "(" + @a.to_s + " and " + @b.to_s + ")"
  end
end

class Exp_Or
  def initialize(a,b)
    @a = a
    @b = b
  end

  def &(arg)
    Exp_And.new(self, arg)
  end

  def |(arg)
    Exp_Or.new(self, arg)
  end

  def to_s
    "(" + @a.to_s + " or " + @b.to_s + ")"
  end
end

class Exp_Implies
  def initialize(a,b)
    @a = a
    @b = $update_store + [b]
    $update_store = []
    puts "a => b " + @a.to_s + " then " + @b.to_s
  end

  def to_s
    "(" + @a.to_s + " implies " + @b.to_s + ")"
  end
end

class Exp_Quant
  def initialize(quant, type, vars, body)
    @quant = quant
    @type = type
    @vars = vars
    @body = body
  end

  def &(arg)
    Exp_And.new(self, arg)
  end

  def |(arg)
    Exp_Or.new(self, arg)
  end

  def to_s
    if @body.is_a? Array then
      body = @body.map{|x| x.to_s}.join("\n")
    else
      body = @body.to_s
    end

    if @quant == :forall then
      puts "THE VARS ARE " + @vars.to_s
      "all " + @vars.join(", ") + ": " + @type.to_s + " { \n" + body + "\n}\n"
    else
      "some " + @vars.join(", ") + ": " + @type.to_s + " { \n" + body + "\n}\n"
    end
  end
end

class Exp_Loop
  def initialize(condition, body)
    @condition = condition
    @body = body
  end

  def to_s
    "while (" + @condition.to_s + ") {\n" + @body.map{|x| x.to_s}.join(";\n") + ";\n}"
  end
end

class Exp_ITE
  def initialize(condition, t, e)
    @condition = condition
    @then_part = t
    @else_part = e
  end

  def to_s
    "ite(" + @condition.to_s + ", " + @then_part.to_s + ", " + @else_part.to_s + ")"
  end
end

class Exp_Parallel
  def initialize(pars)
    @pars = pars
  end

  def to_s
    "parallel(" + @pars.map{|x| x.to_s}.join(", ") + ")"
  end
end

class Exp_SetComp
  def initialize(klass, block)
    @var = Anything.new
    @constraints = block.call(@var)
    puts "CONSTRAINTS ARE: " + @constraints.to_s
    @klass = klass
  end

  def forall(&block)
    res = block.call(@var)
    Exp_Quant.new(:forall, @klass, [@var], Exp_Implies.new(@constraints, res))
  end

  def to_s
    "{" + @var.to_s + " : " + @klass.to_s + " | " + @constraints.to_s + "}"
  end
end

